class UrNotMyDataError(Exception):
    message: str = None

    def __init__(self, msg: str = None, errors: list = None):
        super().__init__(msg)
        self.message = msg or self.message
        self.errors = errors or []

    def __str__(self):
        return '\n' + self.md()

    def md(self, depth=0):
        txt = f'{"  " * depth} - {self.__class__.__name__[:-5]}: {self.message}'
        if self.errors:
            txt = '\n'.join([txt] + [x for x in self.flatten_md(self.errors, depth)])
        return txt

    @classmethod
    def flatten_md(cls, errors, depth=0):
        """ Generate error message at appropriate indent for their nested depth """
        depth += 1
        for el in errors:
            if isinstance(el, list) and not isinstance(el, (str, bytes)):
                yield cls.flatten_md(el, depth)
            elif issubclass(el.__class__, UrNotMyDataError):
                yield el.md(depth)
            else:
                yield UrNotMyDataError(el).md(depth)


class DataError(UrNotMyDataError):
    pass


class DateFormatError(UrNotMyDataError):
    pass


class ConversionError(UrNotMyDataError):
    pass


class CurrencyPatternError(UrNotMyDataError):
    pass


class LengthError(UrNotMyDataError):
    pass


class LeadingZeroError(UrNotMyDataError):
    pass


class InvalidChoiceError(UrNotMyDataError):
    pass


class MissingColumnError(UrNotMyDataError):
    pass


class UnexpectedColumnError(UrNotMyDataError):
    pass


class DuplicateColumnError(UrNotMyDataError):
    pass


class RowLengthError(UrNotMyDataError):
    pass


class NotEnoughFieldsError(UrNotMyDataError):
    pass


class TooManyFieldsError(UrNotMyDataError):
    pass


class ValueComparisonError(UrNotMyDataError):
    pass


class FileUrNotMyDataError(UrNotMyDataError):
    pass


class FilePatternError(UrNotMyDataError):
    pass


class FileError(UrNotMyDataError):
    def __init__(self, file, msg=None, errors: list = None):
        """
        :param file: name of the files that will be reported with the error message.
        :param msg: a custom message to include when reporting errors in this files.
        :param errors: a list of errors contained in this files.
        """
        message = file
        message += f'; {msg}' if msg else ''
        super().__init__(message, errors)


class ColumnError(UrNotMyDataError):
    def __init__(self, msg=None, errors: list = None, **kwargs):
        super().__init__(msg, errors)


class RowError(UrNotMyDataError):
    def __init__(self, index: int, msg=None, errors: list = None, **kwargs):
        """
        :param index: row number (using zero-indexing).
        :param msg: a custom message to include when reporting errors in this row
        :param errors: a list of errors contained in this row
        :param zero_index: specify whether to use zero-indexing when reporting row number.
        If True, the index is reported as provided. If False, the index increased
        by one.
        """
        message = f'{str(index + (0 if kwargs.get("zero_index") else 1))}'
        message += f'; {msg}' if msg else ''
        super().__init__(message, errors)


class CellError(UrNotMyDataError):
    def __init__(self, index: int, msg=None, errors: list = None, **kwargs):
        """
        :param index: column number (using zero-indexing).
        :param msg: a custom message to include when reporting errors in this cell
        :param errors: a list of errors contained in this cell
        :param zero_index: specify whether to use zero-indexing when reporting column number.
        If True, the index is reported as provided. If False, the index increased
        by one.
        """
        message = ''
        offset = 0 if kwargs.get("zero_index") else 1
        if kwargs.get('rix'):
            message = str(kwargs.get('rix') + offset) + ','

        message += f'{str(index + offset)}'
        if kwargs.get("name"):
            message += f' ({kwargs.get("name")})'
        message += f'; {msg}' if msg else ''
        super().__init__(message, errors)


class NullValueError(UrNotMyDataError):
    message = 'value is blank/null'


class NegativeValueError(CellError):
    message = "You're too negative"


class MaxExceededError(UrNotMyDataError):
    pass


class ColumnComparisonError(UrNotMyDataError):
    pass


class RowComparisonError(UrNotMyDataError):
    pass


class CharacterError(UrNotMyDataError):
    pass


class DuplicateValueError(UrNotMyDataError):
    pass


class NoRulesDefinedError(UrNotMyDataError):
    pass



import csv
from pathlib import Path
from typing import Union, Dict

from rumydata import exception as ex
from rumydata import field
from rumydata.base import BaseSubject
from rumydata.rules import file, column as cr, header as hr, row as rr, cell as clr


class Layout(BaseSubject):
    def __init__(self, definition: dict, **kwargs):
        """
        Defines the layout of a tabular files.

        :param definition: dictionary of column names with DataType definitions
        """
        self.skip_header = kwargs.pop('skip_header', False)
        self.empty_row_ok = kwargs.pop('empty_row_ok', False)

        super().__init__(**kwargs)

        self.definition = definition
        self.length = len(definition)
        self.title = kwargs.get('title')

        self.rules.extend([
            hr.ColumnOrder(self),
            hr.NoExtra(self),
            hr.NoDuplicate(self),
            hr.NoMissing(self),
            rr.RowLengthLTE(self.length),
            rr.RowLengthGTE(self.length)
        ])

    def __check__(self, row, rule_type, rix=None):
        if rule_type == hr.Rule and self.skip_header:
            return

        e = super().__check__(row, rule_type=rule_type)

        if e:  # if row errors are found, skip cell checks
            return ex.RowError(rix or -1, errors=e)

        if rule_type == rr.Rule:
            row = dict(zip(self.definition.keys(), row))
            ignore = {k: isinstance(v, field.Ignore) for k, v in self.definition.items()}
            # if empty row is okay, and all fields are either empty, or Ignore class
            if self.empty_row_ok and all([('' if ignore[k] else v) == '' for k, v in row.items()]):
                return

            for cix, (name, val) in enumerate(row.items()):
                t = self.definition[name]
                comp = {k: row[k] for k in t.comparison_columns()}
                check_args = dict(
                    data=(val, comp), rule_type=clr.Rule,
                    rix=rix, cix=cix, name=name
                )
                ce = t.__check__(**check_args)
                if ce:
                    e.append(ce)
            if e:
                return ex.RowError(rix, errors=e)

    def check_header(self, row, rix=0):
        errors = self.__check__(row, rule_type=hr.Rule, rix=rix)
        assert not errors, str(errors)

    def check_row(self, row, rix=-1):
        errors = self.__check__(row, rule_type=rr.Rule, rix=rix)
        assert not errors, str(errors)

    def digest(self):
        return [[f'Name: {k}', *v.digest()] for k, v in self.definition.items()]

    def markdown_digest(self):
        fields = f'# {self.title}' + '\n\n' if self.title else ''
        fields += '\n'.join([
            f' - **{k}**' + ''.join(['\n   - ' + x for x in v.digest()])
            for k, v in self.definition.items()
        ])
        return fields

    def comparison_columns(self):
        compares = set()
        for v in self.definition.values():
            compares.update(v.comparison_columns())
        return compares


class File(BaseSubject):
    def __init__(self, layout: Union[Layout, Dict], max_errors=100, **kwargs):
        self.file_type = kwargs.pop('file_type', 'csv')

        if self.file_type == 'csv':
            x = {x: kwargs.pop(x, None) for x in ['dialect', 'delimiter', 'quotechar']}
            self.csv_kwargs = {k: v for k, v in x.items() if v}
        elif self.file_type == 'excel':
            x = {x: kwargs.pop(x, None) for x in ['sheet']}
            self.excel_kwargs = {k: v for k, v in x.items() if v}
        else:
            raise Exception(f'Invalid file type: {self.file_type}')

        self.skip_rows = kwargs.pop('skip_rows', 0)

        super().__init__(**kwargs)

        self.max_errors = max_errors
        self.layout = Layout(layout) if isinstance(layout, Dict) else layout
        self.rules.extend([
            file.FileExists()
        ])

    def __check__(self, filepath: Union[str, Path], **kwargs):
        p = Path(filepath) if isinstance(filepath, str) else filepath
        e = super().__check__(p, rule_type=file.Rule)  # check files-based rules first
        if e:
            return ex.FileError(file=filepath, errors=e)

        column_cache = {
            k: [] for k, v in self.layout.definition.items()
            if v.has_rule_type(cr.Rule)
        }
        column_cache_map = {
            k: list(self.layout.definition.keys()).index(k)
            for k in column_cache.keys()
        }

        with open(p, newline='') as f:
            if self.file_type == 'csv':
                generator = csv.reader(f, **self.csv_kwargs)

                def row_handler(r):  # handle csv rows as-is
                    return r
            elif self.file_type == 'excel':
                from openpyxl import load_workbook
                wb = load_workbook(p)
                sheet = self.excel_kwargs.get('sheet')
                ws = wb.get_sheet_by_name(sheet) if sheet else wb.active
                generator = ws.values

                def row_handler(r):  # convert values to strings and replace None with empty
                    return [str(x or '') for x in r]

            for rix, row in enumerate(generator):
                if rix < self.skip_rows:
                    continue
                row = row_handler(row)
                rt = hr.Rule if rix == (0 + self.skip_rows) else rr.Rule
                re = self.layout.__check__(row, rule_type=rt, rix=rix)
                if re:
                    e.append(re)
                    if rix == 0:  # if header error present, stop checking rows
                        break
                    if len(e) > self.max_errors:
                        m = f"max of {str(self.max_errors)} row errors exceeded"
                        e.append(ex.MaxExceededError(m))
                        break
                if rix > 0:
                    for k, ix in column_cache_map.items():
                        column_cache[k].append(row[ix])

        for k, v in column_cache.items():
            ce = self.layout.definition[k].__check__(v, rule_type=cr.Rule)
            if ce:
                e.append(ce)
        if e:
            return ex.FileError(file=p.name, errors=e)

    def check(self, file_path):
        errors = self.__check__(file_path)
        assert not errors, str(errors)



from rumydata import exception as ex


class BaseRule:
    """ Base class for defining data type rules """
    exception_class = ex.UrNotMyDataError

    def prepare(self, data) -> tuple:
        return data,

    def evaluator(self):
        """
        :return: a function which expects to evaluate to True if the value
        provided to the function meets the rule.
        """
        return lambda x: False  # default to failing evaluation if not overwritten

    def exception_msg(self):
        """
        :return: a sanitized error message which is specific to the function,
        contains no direct link to the value that was checked.
        """
        return self.exception_class(self.explain())

    def explain(self) -> str:
        """
        :return: an explanation of the rule that is applied
        """
        return "default rule explanation"


class BaseSubject:
    def __init__(self, rules: list = None):
        self.rules = rules or []
        self.descriptors = {}

    def __check__(self, data, rule_type, **kwargs):
        errors = []
        if not self.rules:
            return [ex.NoRulesDefinedError()]

        for r in self.rules:
            # noinspection PyBroadException
            try:
                if issubclass(type(r), rule_type):
                    x = r.prepare(data)
                    e = r.evaluator()(*x)
                    if not e:
                        errors.append(r.exception_msg())
            except Exception as e:  # get type, and rewrite safe message
                errors.append(r.exception_class(
                    f'raised {e.__class__.__name__} while checking if value {r.explain()}')
                )
        return errors

    def __list_errors__(self, value, **kwargs):
        return list(self.flatten_exceptions(self.__check__(value, **kwargs)))

    def __has_error__(self, value, error, **kwargs):
        return error in [x.__class__ for x in self.__list_errors__(value, **kwargs)]

    def digest(self):
        x = [f'{k}: {v}' if v else k for k, v in self.descriptors.items()]
        y = [x.explain() for x in self.rules]
        return x + y

    @classmethod
    def flatten_exceptions(cls, error):
        if isinstance(error, list):
            for el in error:
                yield cls.flatten_exceptions(el)
        elif issubclass(error.__class__, ex.UrNotMyDataError):
            yield error
            for el in error.errors:
                for x in cls.flatten_exceptions(el):
                    yield x
        else:
            yield error



from typing import List, AnyStr

from rumydata import exception as ex
from rumydata.base import BaseRule


class Rule(BaseRule):
    def __init__(self, columns):
        self.definition = columns.definition

    def prepare(self, data: List[AnyStr]) -> tuple:
        return data,


class NoExtra(Rule):
    exception_class = ex.UnexpectedColumnError

    def evaluator(self):
        return lambda x: all([y in self.definition for y in x])

    def explain(self):
        return 'Header row must not have unexpected columns'


class NoMissing(Rule):
    exception_class = ex.MissingColumnError

    def evaluator(self):
        return lambda x: all([y in x for y in self.definition])

    def explain(self) -> str:
        return 'Header row must not be missing any expected columns'


class NoDuplicate(Rule):
    exception_class = ex.DuplicateColumnError

    def evaluator(self):
        return lambda x: len(x) == len(set(x))

    def explain(self):
        return 'Header row must not contain duplicate values'


class ColumnOrder(Rule):
    exception_class = ex.DataError

    def evaluator(self):
        return lambda x: x == list(self.definition)

    def explain(self):
        return 'Header row must explicitly match order of definition'



from typing import Union, Tuple, AnyStr, Dict

from rumydata import exception as ex
from rumydata.base import BaseSubject
from rumydata.rules import cell as clr, column as cr

__all__ = [
    'Field', 'Text', 'Date', 'Currency', 'Digit', 'Integer', 'Choice'
]


class Field(BaseSubject):
    def __init__(self, nullable=False, rules: list = None):
        super().__init__(rules)
        self.nullable = nullable

        if not self.nullable:
            self.rules.append(clr.NotNull())

    def __check__(self, data, cix=-1, rule_type=None, **kwargs):
        # if data is nullable and value is empty, skip all checks
        if self.nullable and issubclass(rule_type, clr.Rule) and data == '':
            pass
        else:
            e = super().__check__(data, rule_type=rule_type)
            if e:
                return ex.CellError(cix, errors=e, **kwargs)

    def check_cell(self, value: Union[AnyStr, Tuple[AnyStr, Dict]], **kwargs):
        errors = self.__check__(value, rule_type=clr.Rule, **kwargs)
        assert not errors, str(errors)

    def check_column(self, values: list, **kwargs):
        errors = self.__check__(values, rule_type=cr.Rule, **kwargs)
        assert not errors, str(errors)

    def digest(self):
        dig = super().digest()
        if self.nullable:
            dig.append('Nullable')
        return dig

    def comparison_columns(self):
        compares = set()
        for r in self.rules:
            if issubclass(type(r), clr.ColumnComparisonRule):
                compares.add(r.compare_to)
        return compares

    def has_rule_type(self, rule_type):
        return any([issubclass(type(r), rule_type) for r in self.rules])


class Ignore(Field):
    """
    Ignore the values found in this field. This will cause other dependencies (like row checks) to
    treat any values in this field as if they were empty. While the check method still exists, this
    class overwrites those inherited methods to intentionally have no effect.
    """

    # noinspection PyMissingConstructor
    def __init__(self):
        self.rules = []
        self.descriptors = {}

    def __check__(self, *args, **kwargs):
        pass


class Text(Field):
    def __init__(self, max_length, min_length=None, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'String'
        self.descriptors['Max Length'] = f'{str(max_length)} characters'

        self.rules.append(clr.MaxChar(max_length))

        if min_length:
            self.descriptors['Min Length'] = f'{str(min_length)} characters'
            self.rules.append(clr.MinChar(min_length))


class Date(Field):
    def __init__(self, min_date: str = None, max_date: str = None, **kwargs):
        rule_kwargs = {
            'truncate_time': kwargs.pop('truncate_time', False)
        }
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Date'
        self.descriptors['Format'] = 'YYYY-MM-DD'

        self.rules.append(clr.CanBeDateIso(**rule_kwargs))

        if max_date:
            self.descriptors['Max Date'] = f'{max_date}'
            self.rules.append(clr.DateLTE(max_date, **rule_kwargs))

        if min_date:
            self.descriptors['Min Date'] = f'{min_date}'
            self.rules.append(clr.DateGTE(min_date, **rule_kwargs))


class Currency(Field):
    def __init__(self, significant_digits: int, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Numeric'
        self.descriptors['Format'] = f'{"9" * (significant_digits - 2)}.99'
        self.descriptors['Max Length'] = f'{str(significant_digits)} digits'

        self.rules.append(clr.MaxDigit(significant_digits))
        self.rules.append(clr.NumericDecimals())


class Digit(Field):
    def __init__(self, max_length, min_length=None, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Numeric'
        self.descriptors['Format'] = f'{"0" * max_length}'
        self.descriptors['Max Length'] = f'{str(max_length)} digits'

        self.rules.append(clr.OnlyNumbers())
        self.rules.append(clr.MaxChar(max_length))

        if min_length:
            self.descriptors['Min Length'] = f'{str(min_length)} digits'
            self.rules.append(clr.MinChar(min_length))


class Integer(Field):
    def __init__(self, max_length, min_length=None, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Numeric'
        self.descriptors['Format'] = f'{"9" * max_length}'
        self.descriptors['Max Length'] = f'{str(max_length)} digits'

        self.rules.append(clr.CanBeInteger())
        self.rules.append(clr.NoLeadingZero())
        self.rules.append(clr.MaxDigit(max_length))

        if min_length:
            self.descriptors['Min Length'] = f'{str(max_length)} digits'
            self.rules.append(clr.MinDigit(min_length))


class Choice(Field):
    def __init__(self, valid_values: list, case_insensitive=False, **kwargs):
        super().__init__(**kwargs)

        self.descriptors['Type'] = 'Choice'
        self.descriptors['Choices'] = ','.join(valid_values)
        self.rules.append(
            clr.Choice(valid_values, case_insensitive=case_insensitive)
        )



from typing import List, AnyStr

from rumydata import exception as ex
from rumydata.base import BaseRule


class Rule(BaseRule):
    exception_class = ex.RowComparisonError

    def prepare(self, data: List[AnyStr]) -> tuple:
        return data,


class RowLengthLTE(Rule):
    exception_class = ex.RowLengthError

    def __init__(self, columns_length):
        self.columns_length = columns_length

    def evaluator(self):
        return lambda x: len(x) <= self.columns_length

    def explain(self) -> str:
        return f'row length must be equal to {str(self.columns_length)}, not greater'


class RowLengthGTE(Rule):
    exception_class = ex.RowLengthError

    def __init__(self, columns_length):
        self.columns_length = columns_length

    def evaluator(self):
        return lambda x: len(x) >= self.columns_length

    def explain(self) -> str:
        return f'row length must be equal to {str(self.columns_length)}, not less'



"""
This is a giant dumpster fire. It needs to be broken apart into a set of more
sensible scripts.
"""

import csv
import tempfile
import uuid
from datetime import datetime as dt
from pathlib import Path

import pytest
from openpyxl import Workbook

import rumydata.file
import rumydata.rules.cell
from rumydata import exception as ex
from rumydata import field
from rumydata.file import File
from rumydata.rules import header as hr, row as rr, column as cr


@pytest.fixture()
def basic() -> dict:
    return {
        'col1': rumydata.field.Text(1),
        'col2': rumydata.field.Integer(1),
        'col3': rumydata.field.Date(),
        'col4': field.Choice(['X', 'Y', 'Z'])
    }


@pytest.fixture()
def minimal_layout():
    return rumydata.file.Layout(dict(x=rumydata.field.Digit(1)))


@pytest.fixture()
def tmpdir():
    with tempfile.TemporaryDirectory() as d:
        yield Path(d)


def write_row(directory, columns: rumydata.file.Layout, row, rows=False):
    p = Path(directory, str(uuid.uuid4()))
    with p.open('w', newline='') as o:
        writer = csv.writer(o)
        writer.writerow(list(columns.definition))
        if rows:
            for r in row:
                writer.writerow(r)
        else:
            writer.writerow(row)
    return p


@pytest.fixture()
def basic_good(tmpdir):
    p = Path(tmpdir, 'good.csv')
    with p.open('w', newline='') as o:
        writer = csv.writer(o)
        writer.writerow(['col1', 'col2', 'col3', 'col4'])
        writer.writerow(['A', '1', '2020-01-01', 'X'])
    yield p.as_posix()


@pytest.fixture()
def basic_good_excel(tmpdir):
    p = Path(tmpdir, 'good.xlsx')
    wb = Workbook()
    ws = wb.active
    ws.append(['col1', 'col2', 'col3', 'col4'])
    ws.append(['A', '1', '2020-01-01', 'X'])
    wb.save(p)
    yield p.as_posix()


@pytest.fixture()
def basic_row_skip_good(tmpdir):
    p = Path(tmpdir, 'good.csv')
    with p.open('w', newline='') as o:
        writer = csv.writer(o)
        writer.writerow(['garbage'])
        writer.writerow(['garbage'])
        writer.writerow(['col1', 'col2', 'col3', 'col4'])
        writer.writerow(['A', '1', '2020-01-01', 'X'])
    yield p.as_posix()


@pytest.fixture()
def readme_layout():
    return {
        'col1': field.Text(8),
        'col2': field.Choice(['x', 'y', 'z'], nullable=True),
        'col3': field.Integer(1)
    }


@pytest.fixture()
def readme_data(tmpdir):
    p = Path(tmpdir, 'bobs_data.csv')
    p.write_text('\n'.join([
        "col1,col2,col3",
        "abc,x,-1",
        "def,,0",
        "ghi,a,1"
    ]))
    yield p.as_posix()


def minimal_bad(rows, directory):
    p = Path(directory, str(uuid.uuid4()))
    with p.open('w') as f:
        w = csv.writer(f)
        w.writerow(['x'])
        for i in range(rows):
            w.writerow(['12'])
    return p


def empty_rows(rows, directory):
    p = Path(directory, str(uuid.uuid4()))
    with p.open('w') as f:
        w = csv.writer(f)
        w.writerow(['x'])
        for i in range(rows):
            w.writerow('')
    return p


def test_file_not_exists(basic):
    assert File(rumydata.file.Layout(basic)).__has_error__('abc123.csv', FileNotFoundError)


@pytest.mark.parametrize('value,kwargs', [
    ('x', dict(max_length=1)),
    ('x', dict(max_length=2)),
    ('', dict(max_length=1, nullable=True)),
    ('', dict(max_length=1, min_length=1, nullable=True))
])
def test_text_good(value, kwargs):
    assert not rumydata.field.Text(**kwargs).check_cell(value)


@pytest.mark.parametrize('value,kwargs,err', [
    ('', dict(max_length=1), ex.NullValueError),
    ('xxx', dict(max_length=2), ex.LengthError),
    ('x', dict(max_length=80, min_length=2), ex.LengthError),
])
def test_text_bad(value, kwargs, err):
    assert rumydata.field.Text(**kwargs).__has_error__(value, err)


@pytest.mark.parametrize('value,kwargs', [
    ('2020-01-01', {}),
    ('', dict(nullable=True)),
    ('2020-01-01', dict(max_date='2020-01-01')),
    ('2020-01-01', dict(max_date='2020-01-02')),
    ('2020-01-01', dict(min_date='2020-01-01', max_date='2020-01-02')),
    ('2020-01-01 00:00:00', dict(truncate_time=True))
])
def test_date_good(value, kwargs):
    assert not field.Date(**kwargs).check_cell(value)


@pytest.mark.parametrize('value,err,kwargs', [
    ('', ex.NullValueError, {}),
    ('20200101', ex.ConversionError, {}),
    ('9999-99-99', ex.ConversionError, {}),
    ('2020-01-01', ex.ValueComparisonError, dict(min_date='2020-01-02')),
    ('2020-01-02', ex.ValueComparisonError, dict(max_date='2020-01-01')),
    ('2020-01-01', ex.ValueComparisonError, dict(min_date='2020-01-02', max_date='2020-01-03')),
    ('2020-01-05', ex.ValueComparisonError, dict(min_date='2020-01-02', max_date='2020-01-03')),
    ('2020-01-01 00:00:00', ex.ConversionError, dict(truncate_time=True)),
    ('2020-01-01 00:00:01', ex.ConversionError, dict(truncate_time=False))
])
def test_date_bad(value, err, kwargs):
    assert rumydata.field.Date(**kwargs).__has_error__(value, err)


@pytest.mark.parametrize('value,sig_dig,kwargs', [
    ('123.45', 5, {}),
    ('123.00', 5, {}),
    ('123.0', 5, {}),
    ('123', 5, {}),
    ('', 1, dict(nullable=True)),
    ('-0.01', 3, dict(rules=[rumydata.rules.cell.NumericLT(0)])),
    ('0', 3, dict(rules=[rumydata.rules.cell.NumericLTE(0)])),
    ('0.00', 3, dict(rules=[rumydata.rules.cell.NumericET(0)])),
    ('0', 3, dict(rules=[rumydata.rules.cell.NumericGTE(0)])),
    ('0.01', 3, dict(rules=[rumydata.rules.cell.NumericGT(0)])),
])
def test_currency_good(value, sig_dig, kwargs):
    assert not rumydata.field.Currency(sig_dig, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,sig_dig,rules,err', [
    ('0.00', 5, [rumydata.rules.cell.NumericLT(0)], ex.ValueComparisonError),
    ('0.01', 5, [rumydata.rules.cell.NumericLTE(0)], ex.ValueComparisonError),
    ('0.01', 5, [rumydata.rules.cell.NumericET(0)], ex.ValueComparisonError),
    ('-0.01', 5, [rumydata.rules.cell.NumericGTE(0)], ex.ValueComparisonError),
    ('-0.01', 5, [rumydata.rules.cell.NumericGT(0)], ex.ValueComparisonError),
    ('', 5, [], ex.NullValueError),
    ('123.45', 4, [], ex.LengthError),
    ('123.', 4, [], ex.CurrencyPatternError),
    ('123.456', 4, [], ex.CurrencyPatternError)
])
def test_currency_bad(value, sig_dig, rules, err):
    assert rumydata.field.Currency(sig_dig, rules=rules).__has_error__(value, err)


@pytest.mark.parametrize('value,max_length, kwargs', [
    ('1', 3, {}),
    ('12', 3, {}),
    ('123', 3, {}),
    ('12', 2, dict(min_length=2)),
    ('123', 3, dict(min_length=2))
])
def test_digit_good(value, max_length, kwargs):
    assert not rumydata.field.Digit(max_length, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,max_length,err,kwargs', [
    ('-123', 3, ex.CharacterError, {}),
    ('-123', 3, ex.LengthError, {}),
    ('1', 2, ex.LengthError, dict(min_length=2)),
    ('5', 3, ex.LengthError, dict(min_length=2))
])
def test_digit_bad(value, max_length, err, kwargs):
    assert rumydata.field.Digit(max_length, **kwargs).__has_error__(value, err)


@pytest.mark.parametrize('value,max_length,kwargs', [
    ('-1', 1, {}),
    ('0', 1, {}),
    ('1', 1, {}),
    ('1', 2, {}),
    ('11', 2, dict(min_length=2)),
    ('', 1, dict(nullable=True)),
    ('-1', 1, dict(rules=[rumydata.rules.cell.NumericLT(0)])),
    ('-1', 1, dict(rules=[rumydata.rules.cell.NumericLTE(0)])),
    ('0', 1, dict(rules=[rumydata.rules.cell.NumericLTE(0)])),
    ('0', 1, dict(rules=[rumydata.rules.cell.NumericET(0)])),
    ('0', 1, dict(rules=[rumydata.rules.cell.NumericGTE(0)])),
    ('1', 1, dict(rules=[rumydata.rules.cell.NumericGTE(0)])),
    ('1', 1, dict(rules=[rumydata.rules.cell.NumericGT(0)]))
])
def test_integer_good(value, max_length, kwargs):
    assert not rumydata.field.Integer(max_length, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,max_length,kwargs,err', [
    ('0', 1, dict(rules=[rumydata.rules.cell.NumericLT(0)]), ex.ValueComparisonError),
    ('1', 1, dict(rules=[rumydata.rules.cell.NumericLTE(0)]), ex.ValueComparisonError),
    ('1', 1, dict(rules=[rumydata.rules.cell.NumericET(0)]), ex.ValueComparisonError),
    ('-1', 1, dict(rules=[rumydata.rules.cell.NumericGTE(0)]), ex.ValueComparisonError),
    ('0', 1, dict(rules=[rumydata.rules.cell.NumericGT(0)]), ex.ValueComparisonError),
    ('', 1, {}, ex.NullValueError),
    ('1', 2, dict(min_length=2), ex.LengthError),
    ('111', 2, {}, ex.LengthError),
    ('00', 2, {}, ex.LeadingZeroError),
    ('01', 2, {}, ex.LeadingZeroError)
])
def test_integer_bad(value, max_length, kwargs, err):
    assert rumydata.field.Integer(max_length, **kwargs).__has_error__(value, err)


@pytest.mark.parametrize('value,choices,kwargs', [
    ('x', ['x'], {}),
    ('x', ['x', 'y'], {}),
    ('y', ['x', 'y'], {}),
    ('', ['x'], dict(nullable=True)),
    ('X', ['x'], dict(case_insensitive=True)),
    ('x', ['X'], dict(case_insensitive=True))
])
def test_choice_good(value, choices, kwargs):
    assert not field.Choice(choices, **kwargs).check_cell(value)


@pytest.mark.parametrize('value,choices,kwargs,err', [
    ('', ['x'], {}, ex.NullValueError),
    ('x', ['z'], {}, ex.InvalidChoiceError)
])
def test_choice_bad(value, choices, kwargs, err):
    assert field.Choice(choices, **kwargs).__has_error__(value, err)


def test_row_good(basic):
    assert not rumydata.file.Layout(basic).check_row(['1', '2', '2020-01-01', 'X'])


@pytest.mark.parametrize('value,err', [
    ([1, 2, 3, 4, 5], ex.RowLengthError),
    ([1, 2, 3], ex.RowLengthError)
])
def test_row_bad(basic, value, err):
    assert rumydata.file.Layout(basic).__has_error__(value, err, rule_type=rr.Rule)


def test_header_good(basic):
    assert not rumydata.file.Layout(basic).check_header(['col1', 'col2', 'col3', 'col4'])


@pytest.mark.parametrize('value,err', [
    (['col1', 'col2'], ex.MissingColumnError),
    (['col1', 'col2', 'col2'], ex.DuplicateColumnError),
    (['col1', 'col2', 'col5'], ex.UnexpectedColumnError)
])
def test_header_bad(basic, value, err):
    assert rumydata.file.Layout(basic).__has_error__(value, err, rule_type=hr.Rule)


def test_header_skip(basic):
    assert not rumydata.file.Layout(basic, skip_header=True).check_header(['col1', 'col2', 'col4'])


def test_file_good(basic_good, basic):
    assert not File(rumydata.file.Layout(basic)).check(basic_good)


def test_file_excel_good(basic_good_excel, basic):
    assert not File(rumydata.file.Layout(basic), file_type='excel').check(basic_good_excel)


def test_file_row_skip_good(basic_row_skip_good, basic):
    assert not File(rumydata.file.Layout(basic), skip_rows=2).check(basic_row_skip_good)


def test_layout_good(basic, basic_good):
    assert not File(rumydata.file.Layout(basic)).check(basic_good)


def test_readme_example(readme_layout, readme_data):
    assert File(rumydata.file.Layout(readme_layout)).__has_error__(readme_data, ex.InvalidChoiceError)


@pytest.mark.parametrize('rows,me', [
    (1, 0),
    (2, 1),
    (101, 100),
    (int(1e5), 100),
])
def test_has_max_error(tmpdir, rows, me):
    fields = rumydata.file.Layout({'x': field.Field()})
    file = empty_rows(rows, tmpdir)
    assert File(fields, max_errors=me).__has_error__(file, ex.MaxExceededError)


@pytest.mark.parametrize('rows,me', [
    (1, 1),
    (99, 100),
    (100, 100),
    (100, int(1e5)),
])
def test_missing_max_error(tmpdir, rows, me):
    fields = rumydata.file.Layout({'x': field.Field()})
    file = empty_rows(rows, tmpdir)
    assert not File(fields, max_errors=me).__has_error__(file, ex.MaxExceededError)


@pytest.mark.parametrize('value,func,assertion', [
    ('2', lambda x: int(x) % 2 == 0, "must be an even number"),
    ('1', lambda x: int(x) % 2 == 1, "must be an odd number"),
    ('2020-01-01', lambda x: dt.fromisoformat(x), "must be an isodate"),
])
def test_static_rules_good(value, func, assertion):
    x = field.Field(rules=[rumydata.rules.cell.make_static_cell_rule(func, assertion)])
    assert not x.check_cell(value)


@pytest.mark.parametrize('value,func,assertion,kwargs', [
    ('2', lambda x: int(x) % 2 == 1, "must be an even number", {}),
    ('1', lambda x: int(x) % 2 == 0, "must be an odd number", {}),
    ('2020-00-01', lambda x: dt.fromisoformat(x), "must be an isodate", {}),
    ('', lambda x: 1 / 0, "custom exception", dict(exception=ZeroDivisionError))
])
def test_static_rules_bad(value, func, assertion, kwargs):
    x = field.Field(rules=[rumydata.rules.cell.make_static_cell_rule(func, assertion, **kwargs)])
    assert x.__has_error__(value, kwargs.get('exception', ex.UrNotMyDataError))


def test_column_compare_rule_good():
    x = field.Field(rules=[rumydata.rules.cell.GreaterThanColumn('x')])
    assert not x.check_cell(('1', {'x': '0'}))


def test_column_compare_rule_bad():
    x = field.Field(rules=[rumydata.rules.cell.GreaterThanColumn('x')])
    assert x.__has_error__('1', compare={'x': '1'}, error=ex.ColumnComparisonError)


def test_column_compare_row_good():
    fields = rumydata.file.Layout({
        'a': rumydata.field.Integer(1, rules=[rumydata.rules.cell.GreaterThanColumn('b')]),
        'b': rumydata.field.Integer(1)
    })
    assert not fields.check_row(['3', '2'])


@pytest.mark.parametrize('compare_rule,row', [
    (rumydata.rules.cell.GreaterThanColumn('x'), ['2', '3']),
])
def test_column_compare_file_good(tmpdir, compare_rule, row):
    cols = rumydata.file.Layout({'x': field.Field(), 'y': field.Field(rules=[compare_rule])})
    assert not File(cols).check(write_row(tmpdir, cols, row))


@pytest.mark.parametrize('compare_rule,row', [
    (rumydata.rules.cell.GreaterThanColumn('x'), ['1', '1']),
])
def test_column_compare_file_bad(tmpdir, compare_rule, row):
    cols = rumydata.file.Layout({'x': field.Field(), 'y': field.Field(rules=[compare_rule])})
    assert File(cols).__has_error__(write_row(tmpdir, cols, row), ex.ColumnComparisonError)


def test_unique_bad(tmpdir):
    cols = rumydata.file.Layout({'x': field.Field(rules=[cr.Unique()])})
    f = write_row(tmpdir, cols, [['1'], ['1'], ['1']], rows=True)
    assert File(cols).__has_error__(f, ex.DuplicateValueError)


def test_unique_good(tmpdir):
    cols = rumydata.file.Layout({'x': field.Field(rules=[cr.Unique()])})
    f = write_row(tmpdir, cols, [['1'], ['2'], ['3']], rows=True)
    assert not File(cols).check(f)


@pytest.mark.parametrize('row,kwargs', [
    (['1', '1'], {}),
    (['1', '1'], dict(empty_row_ok=False)),
    (['', ''], dict(empty_row_ok=True))
])
def test_empty_row_good(row, kwargs):
    lay = rumydata.file.Layout({'x': field.Integer(1), 'y': field.Integer(2)}, **kwargs)
    assert not lay.check_row(row)


def test_empty_row_file_good(tmpdir):
    cols = rumydata.file.Layout({'x': field.Field()}, empty_row_ok=True)
    f = write_row(tmpdir, cols, [['1'], ['2'], ['']], rows=True)
    assert not File(cols).check(f)


@pytest.mark.parametrize('cell', [
    '',
    '1',
    '8k;asdfkl;asdf'
])
def test_ignore_cell(cell):
    assert not field.Ignore().check_cell(cell)


@pytest.mark.parametrize('row', [
    (['1', '']),
    (['', '']),
    (['1', '1'])
])
def test_ignore_row(row):
    """ Test that ignore rows count as empty for the purpose of accepting empty rows """
    lay = rumydata.file.Layout(
        {'x': field.Ignore(), 'y': field.Integer(1)}, empty_row_ok=True
    )
    assert not lay.check_row(row)



import re
from datetime import datetime
from typing import Union, Tuple, Dict, AnyStr, List

from rumydata import exception as ex
from rumydata.base import BaseRule


class Rule(BaseRule):

    def prepare(self, data: Union[AnyStr, Tuple[AnyStr, Dict]]) -> tuple:
        if isinstance(data, str):
            return data,
        else:
            return data[0],


class NotNull(Rule):
    exception_class = ex.NullValueError

    def evaluator(self):
        return lambda x: x != ''

    def exception_msg(self):
        return self.exception_class(self.explain())

    def explain(self):
        return 'cannot be empty/blank'


class ExactChar(Rule):
    exception_class = ex.LengthError

    def __init__(self, exact_length):
        self.exact_length = exact_length

    def evaluator(self):
        return lambda x: len(x) == self.exact_length

    def explain(self):
        return f'must be exactly {str(self.exact_length)} characters'


class MinChar(Rule):
    exception_class = ex.LengthError

    def __init__(self, min_length):
        self.min_length = min_length

    def evaluator(self):
        return lambda x: len(x) >= self.min_length

    def explain(self):
        return f'must be at least {str(self.min_length)} characters'


class MaxChar(Rule):
    exception_class = ex.LengthError

    def __init__(self, max_length):
        self.max_length = max_length

    def evaluator(self):
        return lambda x: len(x) <= self.max_length

    def explain(self):
        return f'must be no more than {str(self.max_length)} characters'


class Choice(Rule):
    exception_class = ex.InvalidChoiceError

    def __init__(self, choices: List[str], case_insensitive=False):
        self.choices = choices
        self.case_insensitive = case_insensitive
        self.eval_choices = [x.lower() for x in choices] if case_insensitive else choices

    def prepare(self, data: Union[str, Tuple[str, Dict]]) -> tuple:
        if self.case_insensitive:
            if isinstance(data, str):
                data = data.lower(),
            else:
                data = data[0].lower(), data[1]

        return super().prepare(data)

    def evaluator(self):
        return lambda x: x in self.eval_choices

    def explain(self):
        return (
            f'must be one of {self.choices}'
            f'{" (case insensitive)" if self.case_insensitive else " (case sensitive)"}'
        )


class MinDigit(Rule):
    """
    Check that count of characters, after removing all non-digits, meets or
    exceeds the specified minimum. Used to evaluate length of significant digits
    in numeric strings that might contain formatting.
    """
    exception_class = ex.LengthError

    def __init__(self, min_length):
        self.min_length = min_length

    def evaluator(self):
        return lambda x: len(re.sub(r'[^\d]', '', x)) >= self.min_length

    def explain(self):
        return f'must have at least {str(self.min_length)} digit characters'


class MaxDigit(Rule):
    """
    Check that count of characters, after removing all non-digits, is less than
    or equal to the specified minimum. Used to evaluate length of significant
    digits in numeric strings that might contain formatting.
    """
    exception_class = ex.LengthError

    def __init__(self, max_length):
        self.max_length = max_length

    def evaluator(self):
        return lambda x: len(re.sub(r'[^\d]', '', x)) <= self.max_length

    def explain(self):
        return f'must have no more than {self.max_length} digit characters'


class OnlyNumbers(Rule):
    exception_class = ex.CharacterError

    def evaluator(self):
        return lambda x: re.fullmatch(r'\d+', x)

    def explain(self):
        return 'must only contain characters 0-9'


class NoLeadingZero(Rule):
    """
    Ensure that there is no leading zero after removing all non-digit characters.
    A lone zero (0) will not raise an error.
    """
    exception_class = ex.LeadingZeroError

    def evaluator(self):
        return lambda x: re.fullmatch(r'(0|([1-9]\d*))', re.sub(r'[^\d]', '', x))

    def explain(self):
        return 'cannot have a leading zero digit'


class CanBeFloat(Rule):
    exception_class = ex.ConversionError

    def evaluator(self):
        return lambda x: isinstance(float(x), float)

    def explain(self):
        return 'can be coerced into a float value'


class CanBeInteger(Rule):
    exception_class = ex.ConversionError

    def evaluator(self):
        return lambda x: isinstance(int(x), int)

    def explain(self):
        return 'can be coerced into an integer value'


class NumericDecimals(Rule):
    exception_class = ex.CurrencyPatternError

    def __init__(self, decimals=2):
        self.decimals = decimals

    def evaluator(self):
        return lambda x: re.fullmatch(r'-?\d+(\.\d{1,2})?', x)

    def explain(self):
        return f'cannot have more than {self.decimals} digits after the decimal point'


class LengthComparison(Rule):
    """
    Base float value comparison class. Requires that the value can be coerced
    to a float value.
    """
    exception_class = ex.ValueComparisonError
    comparison_language = 'N/A'

    def __init__(self, comparison_value):
        self.comparison_value = comparison_value

    def explain(self) -> str:
        return f'{self.comparison_language} {str(self.comparison_value)}'


class LengthGT(LengthComparison):
    """ Length Greater Than comparison """
    comparison_language = 'greater than'

    def evaluator(self):
        return lambda x: len(x) > self.comparison_value


class LengthGTE(LengthComparison):
    """ Length Greater Than or Equal To comparison """
    comparison_language = 'greater than or equal to'

    def evaluator(self):
        return lambda x: len(x) >= self.comparison_value


class LengthET(LengthComparison):
    """ Length Equal To comparison """
    comparison_language = 'equal to'

    def evaluator(self):
        return lambda x: len(x) == self.comparison_value


class LengthLTE(LengthComparison):
    """ Length Less Than or Equal To comparison """
    comparison_language = 'less than or equal to'

    def evaluator(self):
        return lambda x: len(x) <= self.comparison_value


class LengthLT(LengthComparison):
    """ Length Less Than comparison """
    comparison_language = 'less than'

    def evaluator(self):
        return lambda x: len(x) < self.comparison_value


class NumericComparison(Rule):
    """
    Base float value comparison class. Requires that the value can be coerced
    to a float value.
    """
    exception_class = ex.ValueComparisonError
    comparison_language = 'N/A'

    def __init__(self, comparison_value):
        self.comparison_value = comparison_value

    def explain(self) -> str:
        return f'{self.comparison_language} {str(self.comparison_value)}'


class NumericGT(NumericComparison):
    """ Numeric Greater Than comparison """
    comparison_language = 'greater than'

    def evaluator(self):
        return lambda x: float(x) > self.comparison_value


class NumericGTE(NumericComparison):
    """ Numeric Greater Than or Equal To comparison """
    comparison_language = 'greater than or equal to'

    def evaluator(self):
        return lambda x: float(x) >= self.comparison_value


class NumericET(NumericComparison):
    """ Numeric Equal To comparison """
    comparison_language = 'equal to'

    def evaluator(self):
        return lambda x: float(x) == self.comparison_value


class NumericLTE(NumericComparison):
    """ Numeric Less Than or Equal To comparison """
    comparison_language = 'less than or equal to'

    def evaluator(self):
        return lambda x: float(x) <= self.comparison_value


class NumericLT(NumericComparison):
    """ Numeric Less Than comparison """
    comparison_language = 'less than'

    def evaluator(self):
        return lambda x: float(x) < self.comparison_value


class DateRule(Rule):
    exception_class = ex.ConversionError

    def __init__(self, **kwargs):
        self.truncate_time = kwargs.pop('truncate_time', False)
        super().__init__(**kwargs)

    def prepare(self, data: Union[str, Tuple[str, Dict]]) -> tuple:
        if self.truncate_time:
            no_time = ' 00:00:00'
            if isinstance(data, str) and data.endswith(no_time):
                data = data[:-len(no_time)]
            elif data[0].endswith(no_time):
                data = data[0][:-len(no_time)], data[1]

        return super().prepare(data)


class CanBeDateIso(DateRule):
    def evaluator(self):
        return lambda x: isinstance(datetime.strptime(x, '%Y-%m-%d'), datetime)

    def explain(self):
        return 'can be coerced into a ISO-8601 date'


class DateComparison(DateRule):
    """
    Base date value comparison class. Requires that the value can be coerced
    to a date using the specified format for the field.
    """
    exception_class = ex.ValueComparisonError
    comparison_language = 'N/A'

    def __init__(self, comparison_value, date_format='%Y-%m-%d', **kwargs):
        self.date_format = date_format
        self.comparison_value = datetime.strptime(comparison_value, date_format)
        super().__init__(**kwargs)

    def explain(self) -> str:
        return f'{self.comparison_language} {str(self.comparison_value)}'


class DateGT(DateComparison):
    """ Date Greater Than comparison """
    comparison_language = 'greater than'

    def evaluator(self):
        return lambda x: datetime.strptime(x, self.date_format) > self.comparison_value


class DateGTE(DateComparison):
    """ Date Greater Than or Equal To comparison """
    comparison_language = 'greater than or equal to'

    def evaluator(self):
        return lambda x: datetime.strptime(x, self.date_format) >= self.comparison_value


class DateET(DateComparison):
    """ Date Equal To comparison """
    comparison_language = 'equal to'

    def evaluator(self):
        return lambda x: datetime.strptime(x, self.date_format) == self.comparison_value


class DateLTE(DateComparison):
    """ Date Less Than or Equal To comparison """
    comparison_language = 'less than or equal to'

    def evaluator(self):
        return lambda x: datetime.strptime(x, self.date_format) <= self.comparison_value


class DateLT(DateComparison):
    """ Date Less Than comparison """
    comparison_language = 'less than'

    def evaluator(self):
        return lambda x: datetime.strptime(x, self.date_format) < self.comparison_value


def make_static_cell_rule(func, assertion, exception=ex.UrNotMyDataError) -> Rule:
    """
    Return a factory generated Rule class. The function used by the rule must
    directly evaluate a single positional argument (i.e. x, but not x and y).
    Because the Rule cannot be passed a value on initialization, neither the
    evaluator or explain methods in the return class can be dynamic.
    """

    class FactoryRule(Rule):
        exception_class = exception

        def evaluator(self):
            return func

        def explain(self) -> str:
            return assertion

    return FactoryRule()


class ColumnComparisonRule(Rule):
    exception_class = ex.ColumnComparisonError

    def __init__(self, compare_to: str):
        self.compare_to = compare_to

    def prepare(self, data: Tuple[AnyStr, Dict]) -> tuple:
        return data[0], data[1][self.compare_to]


class GreaterThanColumn(ColumnComparisonRule):
    def evaluator(self):
        return lambda x, y: x > y

    def explain(self) -> str:
        return f"must be greater than column '{self.compare_to}'"



import re
from pathlib import Path

import setuptools


def read_version():
    try:
        v = [x for x in Path('rumydata/__init__.py').open() if x.startswith('__version__')][0]
        v = re.match(r"__version__ *= *'(.*?)'\n", v)[1]
        return v
    except Exception as e:
        raise RuntimeError(f"Unable to read version string: {e}")


setuptools.setup(
    name="rumydata",
    version=read_version(),
    author="Christopher Boyd",
    description="A python package for validating expectations of text data, and safely reporting exceptions",
    long_description_content_type="text/markdown",
    url=None,
    long_description=Path('README.md').read_text(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.7",
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'openpyxl'],
    packages=setuptools.find_packages()
)



from typing import List, AnyStr

from rumydata import exception as ex
from rumydata.base import BaseRule


class Rule(BaseRule):

    def prepare(self, data: List[AnyStr]) -> tuple:
        return data,


class Unique(Rule):
    exception_class = ex.DuplicateValueError

    def prepare(self, data: List[AnyStr]) -> tuple:
        return [x for x in data if not x == ''],

    def evaluator(self):
        return lambda x: len(x) == len(set(x))

    def exception_msg(self):
        return self.exception_class(self.explain())

    def explain(self):
        return 'values must be unique'



import re
from pathlib import Path
from typing import Union, List

from rumydata import exception as ex
from rumydata.base import BaseRule


class Rule(BaseRule):

    def prepare(self, data: Union[str, Path]) -> tuple:
        return data,


class FileExists(Rule):
    exception_class = FileNotFoundError

    def evaluator(self):
        return lambda x: Path(x).exists()

    def explain(self) -> str:
        return 'files must exist'


class FileNameMatchesPattern(Rule):
    exception_class = ex.FilePatternError

    def __init__(self, pattern: Union[re.Pattern, List[re.Pattern]]):
        self.patterns = [pattern] if isinstance(pattern, re.Pattern) else pattern

    def evaluator(self):
        return lambda x: any([p.fullmatch(Path(x).name) for p in self.patterns])

    def explain(self) -> str:
        return 'files name must match a pattern provided in the layout'


class FileNameMatchesOnePattern(Rule):
    exception_class = ex.UrNotMyDataError

    def __init__(self, patterns: list):
        self.patterns = patterns

    def evaluator(self):
        return lambda x: sum([
            True if p.fullmatch(Path(x).name) else False for p in self.patterns
        ]) <= 1

    def explain(self) -> str:
        return 'files cannot match multiple patterns provided in the layout'
